@@ -355,14 +355,20 @@ class Application(SingletonConfigurable):
 
     #: the alias map for configurables
     #: Keys might strings or tuples for additional options; single-letter alias accessed like `-v`.
-    #: Values might be like "Class.trait" strings of two-tuples: (Class.trait, help-text).
-    aliases: t.Dict[str, t.Any] = {"log-level": "Application.log_level"}
+    #: Values might be like "Class.trait" strings of two-tuples: (Class.trait, help-text),
+    #  or just the "Class.trait" string, in which case the help text is inferred from the
+    #  corresponding trait
+    aliases: t.Dict[t.Union[str, t.Tuple[str, ...]], t.Union[str, t.Tuple[str, str]]] = {
+        "log-level": "Application.log_level"
+    }
 
     # flags for loading Configurables or store_const style flags
     # flags are loaded from this dict by '--key' flags
     # this must be a dict of two-tuples, the first element being the Config/dict
     # and the second being the help string for the flag
-    flags: t.Dict[str, t.Any] = {
+    flags: t.Dict[
+        t.Union[str, t.Tuple[str, ...]], t.Tuple[t.Union[t.Dict[str, t.Any], Config], str]
+    ] = {
         "debug": (
             {
                 "Application": {
@@ -394,7 +400,7 @@ class Application(SingletonConfigurable):
     # this must be a dict of two-tuples,
     # the first element being the application class/import string
     # and the second being the help string for the subcommand
-    subcommands: t.Union[t.Dict[str, t.Tuple[str, str]], Dict] = Dict()
+    subcommands: t.Union[t.Dict[str, t.Tuple[t.Any, str]], Dict] = Dict()
     # parse_command_line will initialize a subapp, if requested
     subapp = Instance("traitlets.config.application.Application", allow_none=True)
 
@@ -519,6 +525,7 @@ class Application(SingletonConfigurable):
             for c in cls.mro()[:-3]:
                 classdict[c.__name__] = c
 
+        fhelp: t.Optional[str]
         for alias, longname in self.aliases.items():
             try:
                 if isinstance(longname, tuple):
@@ -533,12 +540,13 @@ class Application(SingletonConfigurable):
                 fhelp = cls.class_get_trait_help(trait, helptext=fhelp).splitlines()
 
                 if not isinstance(alias, tuple):
-                    alias = (alias,)  # type:ignore[assignment]
+                    alias = (alias,)
                 alias = sorted(alias, key=len)  # type:ignore[assignment]
                 alias = ", ".join(("--%s" if len(m) > 1 else "-%s") % m for m in alias)
 
                 # reformat first line
-                fhelp[0] = fhelp[0].replace("--" + longname, alias)
+                assert fhelp is not None
+                fhelp[0] = fhelp[0].replace("--" + longname, alias)  # type:ignore
                 yield from fhelp
                 yield indent("Equivalent to: [--%s]" % longname)
             except Exception as ex:
@@ -557,7 +565,7 @@ class Application(SingletonConfigurable):
         for flags, (cfg, fhelp) in self.flags.items():
             try:
                 if not isinstance(flags, tuple):
-                    flags = (flags,)  # type:ignore[assignment]
+                    flags = (flags,)
                 flags = sorted(flags, key=len)  # type:ignore[assignment]
                 flags = ", ".join(("--%s" if len(m) > 1 else "-%s") % m for m in flags)
                 yield flags
@@ -744,7 +752,7 @@ class Application(SingletonConfigurable):
         for alias, longname in self.aliases.items():
             if isinstance(longname, tuple):
                 longname, _ = longname
-            cls, trait = longname.split(".", 1)
+            cls, trait = longname.split(".", 1)  # type:ignore
             children = mro_tree[cls]  # type:ignore[index]
             if len(children) == 1:
                 # exactly one descendent, promote alias
@@ -759,7 +767,7 @@ class Application(SingletonConfigurable):
         flags = {}
         for key, (flagdict, help) in self.flags.items():
             newflag: t.Dict[t.Any, t.Any] = {}
-            for cls, subdict in flagdict.items():
+            for cls, subdict in flagdict.items():  # type:ignore
                 children = mro_tree[cls]  # type:ignore[index]
                 # exactly one descendent, promote flag section
                 if len(children) == 1:
@@ -771,7 +779,7 @@ class Application(SingletonConfigurable):
                     newflag[cls] = subdict
 
             if not isinstance(key, tuple):
-                key = (key,)  # type:ignore[assignment]
+                key = (key,)
             for k in key:
                 flags[k] = (newflag, help)
         return flags, aliases
@@ -779,11 +787,58 @@ class Application(SingletonConfigurable):
     def _create_loader(self, argv, aliases, flags, classes):
         return KVArgParseConfigLoader(argv, aliases, flags, classes=classes, log=self.log)
 
+    @classmethod
+    def _get_sys_argv(cls, check_argcomplete: bool = False) -> t.List[str]:
+        """Get `sys.argv` or equivalent from `argcomplete`
+
+        `argcomplete`'s strategy is to call the python script with no arguments,
+        so ``len(sys.argv) == 1``, and run until the `ArgumentParser` is constructed
+        and determine what completions are available.
+
+        On the other hand, `traitlet`'s subcommand-handling strategy is to check
+        ``sys.argv[1]`` and see if it matches a subcommand, and if so then dynamically
+        load the subcommand app and initialize it with ``sys.argv[1:]``.
+
+        This helper method helps to take the current tokens for `argcomplete` and pass
+        them through as `argv`.
+        """
+        if check_argcomplete and "_ARGCOMPLETE" in os.environ:
+            try:
+                from traitlets.config.argcomplete_config import get_argcomplete_cwords
+
+                cwords = get_argcomplete_cwords()
+                assert cwords is not None
+                return cwords
+            except (ImportError, ModuleNotFoundError):
+                pass
+        return sys.argv
+
+    @classmethod
+    def _handle_argcomplete_for_subcommand(cls):
+        """Helper for `argcomplete` to recognize `traitlets` subcommands
+
+        `argcomplete` does not know that `traitlets` has already consumed subcommands,
+        as it only "sees" the final `argparse.ArgumentParser` that is constructed.
+        (Indeed `KVArgParseConfigLoader` does not get passed subcommands at all currently.)
+        We explicitly manipulate the environment variables used internally by `argcomplete`
+        to get it to skip over the subcommand tokens.
+        """
+        if "_ARGCOMPLETE" not in os.environ:
+            return
+
+        try:
+            from traitlets.config.argcomplete_config import increment_argcomplete_index
+
+            increment_argcomplete_index()
+        except (ImportError, ModuleNotFoundError):
+            pass
+
     @catch_config_error
     def parse_command_line(self, argv=None):
         """Parse the command line arguments."""
         assert not isinstance(argv, str)
-        argv = sys.argv[1:] if argv is None else argv
+        if argv is None:
+            argv = self._get_sys_argv(check_argcomplete=bool(self.subcommands))[1:]
         self.argv = [cast_unicode(arg) for arg in argv]
 
         if argv and argv[0] == "help":
@@ -795,6 +850,7 @@ class Application(SingletonConfigurable):
             subc, subargv = argv[0], argv[1:]
             if re.match(r"^\w(\-?\w)*$", subc) and subc in self.subcommands:
                 # it's a subcommand, and *not* a flag or class parameter
+                self._handle_argcomplete_for_subcommand()
                 return self.initialize_subcommand(subc, subargv)
 
         # Arguments after a '--' argument are for the script IPython may be
@@ -948,7 +1004,7 @@ class Application(SingletonConfigurable):
         """generate default config file from Configurables"""
         lines = ["# Configuration file for %s." % self.name]
         lines.append("")
-        lines.append("c = get_config()  # noqa")
+        lines.append("c = get_config()  #" + "noqa")
         lines.append("")
         classes = self.classes if classes is None else classes
         config_classes = list(self._classes_with_config_traits(classes))
